package at.ac.uibk.gitsearch.service;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.ParseException;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipInputStream;

import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants.AuthoritiesConstantEnum;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class SearchService {

	private final Logger log = LoggerFactory.getLogger(SearchService.class);

	protected final MetaDataRepository metaDataRepository;
	protected final TokenProvider tokenProvider;
	protected final GitLabRepository gitLabRepository;
    protected final GitlabService gitlabService;
	protected final PluginManagementService pluginManagementService;
	protected final ShoppingBasketService shoppingBasketService;

	public SearchService(MetaDataRepository metaDataRepository, TokenProvider tokenProvider, GitLabRepository gitLabRepository, GitlabService gitlabService, PluginManagementService pluginManagementService, ShoppingBasketService shoppingBasketService) {
        this.metaDataRepository = metaDataRepository;
        this.tokenProvider = tokenProvider;
        this.gitLabRepository = gitLabRepository;
        this.gitlabService = gitlabService;
        this.pluginManagementService = pluginManagementService;
        this.shoppingBasketService = shoppingBasketService;
    }

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix, int max) throws IOException {
		return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix, max);
	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param formatPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getFormatsAutoComplete(String formatPrefix, int max) throws IOException {
		return metaDataRepository.getFormatsAutoComplete(formatPrefix, max);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix, int max) throws IOException {
		return metaDataRepository.getCreatorAutoComplete(creatorPrefix, max);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix, int max) throws IOException {
		return metaDataRepository.getContributorAutoComplete(contributorPrefix, max);
	}

	/**
	 * returns all contributor and author autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix, int max)
			throws IOException {
		return metaDataRepository.getContributorCreatorAutoComplete(contributorPrefix, max);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param programmingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix, int max)
			throws IOException {
		return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix, max);
	}

	/**
	 * returns the result page
	 *
	 * @param searchInput the query definition
	 */
	public SearchResultsDTO searchResultPage(SearchInputDTO searchInput) throws IOException {
		final Optional<User> principal = tokenProvider.getCurrentPrincipal();
		log.debug("Searchrequest for {} ", searchInput);

		final SearchResultsDTO pageDetails = metaDataRepository.pageDetails(searchInput, principal);

		pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);
        pageDetails.getSearchResult()
            .forEach(this::filterArtemisActionForUnauthorized);

		return pageDetails;
	}

	public SearchResultsDTO searchResultForAuthorEmail(String searchInput, int length) throws IOException {
		log.debug("Searchrequest for {} ", searchInput);

		final SearchResultsDTO pageDetails = metaDataRepository.searchByEmail(searchInput, length,
				tokenProvider.getCurrentPrincipal());

				pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

		return pageDetails;
	}
	/**
	 * returns the result page
	 *
	 * @param page        the index of the first record to be returned
	 * @param length      the number of records on page
	 */
	public SearchResultsDTO watchListResultPage(Stream<String> exerciseIds, int page, int length) throws IOException {

		final SearchResultsDTO pageDetails = metaDataRepository.getExercisesById(exerciseIds,
				tokenProvider.getCurrentPrincipal(), page, length);

		pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

		return pageDetails;
		// return readTestResults(searchInput.getFulltextQuery(), first, length);
	}

	private void fixImageURL(SearchResultDTO metaData) {

		String image = metaData.getMetadata().getImage();

		try {
			if (image != null) {
				URI url = new URI(image);
				if (url.isAbsolute())
					return;
				String httpUrlToRepo = metaData.getProject().getUrl();
				String baseRepoURL = httpUrlToRepo + "/-/raw/master/";
				final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
				metaData.getMetadata().setImage(resolvedImageUrl.toASCIIString());
				return;
			}

			final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
			final boolean isJava = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA"))
					&& metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
			final boolean isLatexFormat = metaData.getMetadata().getFormat() != null
					&& Arrays.stream(metaData.getMetadata().getFormat()).anyMatch(s -> s.startsWith("latex"));
			final boolean isPython = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
			if (isJava) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/java.png");
				return;
			}
			// depends on code commented out above
//			if (isLatexFormat) {
//				metaData.getMetadata().setImage(baseUrl + "/content/img/latex.png");
//				return;
//			}
			if (isPython) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/python.png");
				return;
			}
			metaData.getMetadata().setImage(baseUrl + "/content/img/gitLab.png");
			return;

		} catch (URISyntaxException e) {
			log.warn("Cannot parse image url {} for {}", metaData.getMetadata().getImage(),
					metaData.getProject().getUrl());
		}
	}

	public File exportExercise(String exerciseId) throws IOException, ParseException {
		return exportExercise( ExerciseId.fromString(exerciseId));
	}
	public File exportExercise(ExerciseId exerciseId) throws IOException {
		try {
			final Optional<User> currentPrincipal = tokenProvider.getCurrentPrincipal();
			final SearchResultDTO exercise = metaDataRepository.getExerciseById(exerciseId,
					currentPrincipal);
			final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
			// first we try with user credentials!
			InputStream repositoryArchive = null;
			String[] filter = null;
			try {
				repositoryArchive = gitLabApi.getRepositoryApi()
						.getRepositoryArchive(exercise.getProject().getProject_id(), "HEAD", "zip");
				// if the previous gitlab requests succeeds, the user has full access to the repository.
				// no filtering is required
				filter = new String[] {};
			} catch (GitLabApiException e) {
				// if the previous gitlab throws an exception, we expect, that this is because of permission problems.
				// we check for public visibility, and construct the filter from the except array.
				log.warn("no access?", e.getMessage());
				if (exercise.getMetadata().getPublicVisibility()!=null) {
					repositoryArchive = gitLabRepository.getAdminGitLabApi().getRepositoryApi()
							.getRepositoryArchive(exercise.getProject().getProject_id(), "HEAD", "zip");
					filter = exercise.getMetadata().getPublicVisibility().getExcept();
				} else { // throw it further
					throw e;
				}

			}
			InputStream inputFile = shoppingBasketService.rePackageGitLabProjectZip(
					new ZipInputStream(repositoryArchive),
					"from project " + exerciseId.getProjectId(),
					filter,
					exerciseId.getPath());
			File file = new File("exercise" + exerciseId.getProjectId() + ".zip");

			return copyInputStreamToFile(inputFile, file);
		} catch (GitLabApiException exception) {
			log.error(exception.getMessage());
			return null;
		}
	}

	private File copyInputStreamToFile(InputStream inputStream, File file) throws IOException {

		// append = false
		try (FileOutputStream outputStream = new FileOutputStream(file, false)) {
			int read;
			byte[] bytes = new byte[8192];
			while ((read = inputStream.read(bytes)) != -1) {
				outputStream.write(bytes, 0, read);
			}
			return file;
		}
	}

	public Boolean hasAccessToExerciseId(String exerciseId) {

		if (tokenProvider.getCurrentPrincipal().isEmpty()) {
			log.warn("Cannot find a principal for for exercise {}", exerciseId);
			return false;
		}
		if (tokenProvider.getCurrentPrincipal().get().getAuthorities()
				.contains(AuthoritiesConstantEnum.ADMIN.getGrantedAuthority())) {
			log.info("Admin can access everything : {}", exerciseId);
			return true; // ADMIN is always allowed
		}
		SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId, tokenProvider.getCurrentPrincipal());
		if (result == null) {
			log.warn("Cannot find exercise for : {}", exerciseId);
			return false;
		}
		return true;
	}

	public Optional<SearchResultDTO> findExerciseById(ExerciseId exerciseId) {
		try {
			SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId.toString(),
					tokenProvider.getCurrentPrincipal());
			return Optional.ofNullable(result);
		} catch (javax.ws.rs.NotFoundException e) {
			log.error("exercise with id {} not found?", exerciseId, e);
			return Optional.empty();
		}
	}

    /**
     * Method used to filter out the Artemis export action for
     * non-authorized users having only partial access to the data.
     *
     * See Metadata entry 'publicVisibility.except'
     * @param result SearchResultDTO to filter the actions from
     * @return the updated SearchResultDTO with filtered actions
     */
    private SearchResultDTO filterArtemisActionForUnauthorized(SearchResultDTO result) {
        if (result.getMetadata().getPublicVisibility() == null || result.getMetadata().getPublicVisibility().getExcept() == null) {
            // results queried without 'publicVisibility' are already authorized
            return result;
        }
        try {
            GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
            gitLabApi.getRepositoryApi().getRepositoryArchive(result.getProject().getProject_id(), "HEAD", "zip");
        } catch (GitLabApiException e) {
            // We expect permissions missing
            result.setSupportedActions(result.getSupportedActions().stream().filter(action -> !action.getPlugin().equals("Artemis Sharing Connector")).collect(Collectors.toList()));
        }
        return result;
    }

}
