package at.ac.uibk.gitsearch.service;

import static at.ac.uibk.gitsearch.web.rest.AccountResourceIT.TEST_USER_LOGIN;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.gitlab4j.api.models.Group;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.MockitoAnnotations;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;

@SpringBootTest(classes = GitsearchApp.class)
@RunWith(SpringJUnit4ClassRunner.class)
@WithMockUser(value = TEST_USER_LOGIN, authorities = "sharing")
public class ExerciseServiceIT {

    @InjectMocks
    @Autowired
    ExerciseService exerciseService;

    @MockBean
    GitlabService gitlabService;

    String token = "testToken";

    ArtemisExerciseInfo artemisExerciseInfo = new ArtemisExerciseInfo();

    @Before
    public void init() throws IOException {
        MockitoAnnotations.initMocks(this);
        artemisExerciseInfo.setMetadataVersion("1");
        artemisExerciseInfo.setLearningResourceType("Artemis");
        artemisExerciseInfo.setTitle("TestTitle");
        artemisExerciseInfo.setLicense("testLicense");
    }

    @After
    public void cleanUp() throws IOException {
        if (new File("tmp-import/", token).exists()) {
            FileUtils.deleteDirectory(new File("tmp-import/" + token));
        }
    }

    @Test
    public void testImportExercise() throws Exception {
        int gitlabGroup = 1;

        File metadata = new File("tmp-import/" + token, "Exercise-Details-something.json");
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        Group group = new Group();
        Repository jgitRepo = FileRepositoryBuilder.create(new File("tmp-import/" + token, ".git"));

        // mocks
        doNothing().when(gitlabService).setJGitDefaultAuth();
        doReturn(group).when(gitlabService).getGroupById(gitlabGroup);
        doReturn("repoUrl").when(gitlabService).createRepository(group, artemisExerciseInfo.getTitle());
        doReturn(jgitRepo).when(gitlabService).checkoutRepo("repoUrl");
        doNothing().when(gitlabService).commitAndPushToRepo(jgitRepo, new File("tmp-import", token));

        exerciseService.importExercise(artemisExerciseInfo, token, gitlabGroup);

        // asserts
        verify(gitlabService, times(1)).createRepository(group, artemisExerciseInfo.getTitle());
        verify(gitlabService, times(1)).commitAndPushToRepo(jgitRepo, new File("tmp-import", token));
    }

    @Test
    public void testApplyExerciseInfoChanges() throws Exception {
        File metadata = new File("tmp-import/" + token, "metadata.yaml");
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();

        exerciseService.applyExerciseInfoChanges(artemisExerciseInfo, token);

        Assert.assertEquals("learningResourceType: \"Artemis\"\n" + "license: \"testLicense\"\n"
                + "metadataVersion: \"1\"\n" + "title: \"TestTitle\"",
                new String(Files.readAllBytes(metadata.toPath())).strip());
    }

    @Test
    public void testGetArtemisExerciseInfo() throws Exception {
        File metadata = new File("tmp-import/" + token, "Exercise-Details-something.json");
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        FileWriter writer = new FileWriter(metadata);
        writer.write("{\n" + "  \"metadataVersion\": \"1\",\n" + "  \"type\": \"programming exercise\",\n"
                + "  \"title\": \"TestTitle\",\n" + "  \"license\": \"testLicense\",\n" + "  \"maxPoints\": 100,\n"
                + "  \"bonusPoints\": 100,\n" + "  \"mode\": \"INDIVIDUAL\",\n" + "  \"staticCodeAnalysis\": false,\n"
                + "  \"allowOfflineIDE\": false,\n" + "  \"allowOnlineEditor\": false,\n"
                + "  \"showTestNamesToStudents\": false,\n" + "  \"sequentialTestRuns\": false\n" + "}");
        writer.close();

        artemisExerciseInfo.setMaxPoints(100f);
        artemisExerciseInfo.setBonusPoints(100f);
        artemisExerciseInfo.setMode(ArtemisExerciseInfo.ExerciseMode.INDIVIDUAL);
        artemisExerciseInfo.setStaticCodeAnalysis(false);
        artemisExerciseInfo.setAllowOfflineIDE(false);
        artemisExerciseInfo.setAllowOnlineEditor(false);
        artemisExerciseInfo.setShowTestNamesToStudents(false);
        artemisExerciseInfo.setSequentialTestRuns(false);
        artemisExerciseInfo.setLearningResourceType(null);
        artemisExerciseInfo.setStructure("atomic");
        List<String> str = new ArrayList<>();
        str.add("artemis");


        artemisExerciseInfo.setFormat(str);

        Assert.assertEquals(artemisExerciseInfo, exerciseService.getArtemisExerciseInfo(token));
    }

    @Test
    public void testGetArtemisExerciseInfo_onlyMetadata() throws Exception {
        File metadata = new File("tmp-import/" + token, "Exercise-Details-something.json");
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        FileWriter writer = new FileWriter(metadata);
        writer.write("{\n" + "  \"metadataVersion\": \"1\",\n" + "  \"type\": \"programming exercise\",\n"
                + "  \"title\": \"TestTitle\",\n" + "  \"license\": \"testLicense\"\n" + "}");
        writer.close();

        artemisExerciseInfo.setLearningResourceType(null);
        artemisExerciseInfo.setStructure("atomic");
        List<String> str = new ArrayList<>();
        str.add("artemis");


        artemisExerciseInfo.setFormat(str);

        Assert.assertEquals(artemisExerciseInfo, exerciseService.getArtemisExerciseInfo(token));
    }

    @Test
    public void testValidate() throws Exception {
        String testApiKeyCorrect = "2c8845a4-b3df-414b-a682-36e2313dc1c0";
        String testApiKeyIncorrect = "wrong-apikey";

        Assert.assertTrue(exerciseService.validate(testApiKeyCorrect));
        Assert.assertFalse(exerciseService.validate(testApiKeyIncorrect));

    }
}
